/* $Id: rtcm.c 4364 2007-05-31 08:43:23Z esr $ */
/*****************************************************************************

This is a decoder for RTCM-104, an obscure and complicated serial
protocol used for broadcasting pseudorange corrections from
differential-GPS reference stations.  The applicable
standard is

RTCM RECOMMENDED STANDARDS FOR DIFFERENTIAL NAVSTAR GPS SERVICE,
RTCM PAPER 194-93/SC 104-STD

Ordering instructions are accessible from <http://www.rtcm.org/>
under "Publications".  This describes version 2.1 of the RTCM specification.

Also applicable is ITU-R M.823: "Technical characteristics of
differential transmissions for global navigation satellite systems
from maritime radio beacons in the frequency band 283.5 - 315 kHz in
region 1 and 285 - 325 kHz in regions 2 & 3."

The RTCM protocol uses as a transport layer the GPS satellite downlink
protocol described in IS-GPS-200, the Navstar GPS Interface
Specification.  This code relies on the lower-level packet-assembly
code for that protocol in isgps.c.

The lower layer's job is done when it has assembled a message of up to
33 words of clean parity-checked data.  At this point this upper layer
takes over.  struct rtcm_msg_t is overlaid on the buffer and the bitfields
are used to extract pieces of it.  Those pieces are copied and (where
necessary) reassembled into a struct rtcm_t.

This code and the contents of isgps.c are evolved from code by Wolgang
Rupprecht.  Wolfgang's decoder was loosely based on one written by
John Sager in 1999 (in particular the dump function emits a close
descendant of Sager's dump format).  Here are John Sager's original
notes:

The RTCM decoder prints a legible representation of the input data.
The RTCM SC-104 specification is copyrighted, so I cannot
quote it - in fact, I have never read it! Most of the information
used to develop the decoder came from publication ITU-R M.823.
This is a specification of the data transmitted from LF DGPS
beacons in the 300kHz band. M.823 contains most of those parts of
RTCM SC-104 directly relevant to the air interface (there
are one or two annoying and vital omissions!). Information
about the serial interface format was gleaned from studying
the output of a beacon receiver test program made available on
Starlink's website.

*****************************************************************************/

#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <stdio.h>
#include <math.h> 		/* for round() */
#include <arpa/inet.h>
#include <ctype.h>

#include "gpsd_config.h"
#include "gpsd.h"
#include "rtcm.h"

#ifdef RTCM104_ENABLE

#define PREAMBLE_PATTERN 0x66
#define BIG_ENDIAN_PROTOCOL
#include "bits.h"

static unsigned int tx_speed[] = { 25, 50, 100, 110, 150, 200, 250, 300 };

struct ambiguity_t {
        bool    l1_valid;  
        bool    l2_valid;  
        long l1_cycles;
        long l2_cycles;
	unsigned int l1_start;
	unsigned int l2_start;
};

/* Set initial L1 ambiguity values for all 32 GPS satellites for RTCM 3.0 L1 tracking issues */
/* set to number of satellites plus 1 to keep indexing 1 based instead of 0 based (ie. I'm lazy) */
static struct ambiguity_t phase_ambiguity[] = {
	{0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0},
	{0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0},
	{0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0},
	{0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0},
	{0, 0, 0}
};


#define DIMENSION(a) (unsigned)(sizeof(a)/sizeof(a[0]))

void rtcm_unpack(/*@out@*/struct rtcm_t *tp, char *buf)
/* break out the raw bits into the content fields */
{
    int len;
    unsigned int n, w;
    struct rtcm_msg_t *msg = (struct rtcm_msg_t *)buf;

    long ltmp;
    unsigned long ultmp;
    unsigned long long u38tmp;
    char ctmp[BUFSIZ];
    unsigned short s;

    tp->type = msg->w1.msgtype;
    tp->length = msg->w2.frmlen;
    tp->zcount = msg->w2.zcnt * ZCOUNT_SCALE;
    tp->refstaid = msg->w1.refstaid;
    tp->seqnum = msg->w2.sqnum;
    tp->stathlth = msg->w2.stathlth;

    len = (int)tp->length;
    n = 0;
    switch (tp->type) {
    case 1:
    case 9:
	{
	    struct b_correction_t    *m = &msg->msg_type.type1.corrections[0];

	    while (len >= 0) {
		if (len >= 2) {
		    tp->msg_data.ranges.sat[n].ident      = m->w3.satident1;
		    tp->msg_data.ranges.sat[n].udre       = m->w3.udre1;
		    tp->msg_data.ranges.sat[n].issuedata  = m->w4.issuedata1;
		    tp->msg_data.ranges.sat[n].rangerr    = m->w3.pc1 * 
			(m->w3.scale1 ? PCLARGE : PCSMALL);
		    tp->msg_data.ranges.sat[n].rangerate  = m->w4.rangerate1 * 
					(m->w3.scale1 ? RRLARGE : RRSMALL);
		    n++;
		}
		if (len >= 4) {
		    tp->msg_data.ranges.sat[n].ident      = m->w4.satident2;
		    tp->msg_data.ranges.sat[n].udre       = m->w4.udre2;
		    tp->msg_data.ranges.sat[n].issuedata  = m->w6.issuedata2;
		    tp->msg_data.ranges.sat[n].rangerr    = m->w5.pc2 * 
			(m->w4.scale2 ? PCLARGE : PCSMALL);
		    tp->msg_data.ranges.sat[n].rangerate  = m->w5.rangerate2 * 
			(m->w4.scale2 ? RRLARGE : RRSMALL);
		    n++;
		}
		if (len >= 5) {
		    tp->msg_data.ranges.sat[n].ident       = m->w6.satident3;
		    tp->msg_data.ranges.sat[n].udre        = m->w6.udre3;
		    tp->msg_data.ranges.sat[n].issuedata   = m->w7.issuedata3;
		    /*@ -shiftimplementation @*/
		    tp->msg_data.ranges.sat[n].rangerr     = ((m->w6.pc3_h<<8)|(m->w7.pc3_l)) *
					(m->w6.scale3 ? PCLARGE : PCSMALL);
		    tp->msg_data.ranges.sat[n].rangerate   = m->w7.rangerate3 * 
					(m->w6.scale3 ? RRLARGE : RRSMALL);
		    /*@ +shiftimplementation @*/
		    n++;
		}
		len -= 5;
		m++;
	    }
	    tp->msg_data.ranges.nentries = n;
	}
	break;
    case 3:
        {
	    struct rtcm_msg3    *m = &msg->msg_type.type3;

	    if ((tp->msg_data.ecef.valid = len >= 4)) {
		tp->msg_data.ecef.x = ((m->w3.x_h<<8)|(m->w4.x_l))*XYZ_SCALE;
		tp->msg_data.ecef.y = ((m->w4.y_h<<16)|(m->w5.y_l))*XYZ_SCALE;
		tp->msg_data.ecef.z = ((m->w5.z_h<<24)|(m->w6.z_l))*XYZ_SCALE;
	    }
	}
	break;
    case 4:
	if ((tp->msg_data.reference.valid = len >= 2)){
	    struct rtcm_msg4    *m = &msg->msg_type.type4;

	    tp->msg_data.reference.system =
		    (m->w3.dgnss==0) ? gps :
			    ((m->w3.dgnss==1) ? glonass : unknown);
	    tp->msg_data.reference.sense = (m->w3.dat != 0) ? global : local;
	    if (m->w3.datum_alpha_char1){
		tp->msg_data.reference.datum[n++] = (char)(m->w3.datum_alpha_char1);
	    }
	    if (m->w3.datum_alpha_char2){
		tp->msg_data.reference.datum[n++] = (char)(m->w3.datum_alpha_char2);
	    }
	    if (m->w4.datum_sub_div_char1){
		tp->msg_data.reference.datum[n++] = (char)(m->w4.datum_sub_div_char1);
	    }
	    if (m->w4.datum_sub_div_char2){
		tp->msg_data.reference.datum[n++] = (char)(m->w4.datum_sub_div_char2);
	    }
	    if (m->w4.datum_sub_div_char3){
		tp->msg_data.reference.datum[n++] = (char)(m->w4.datum_sub_div_char3);
	    }
	    tp->msg_data.reference.datum[n++] = '\0';
	    if (len >= 4) {
		tp->msg_data.reference.dx = m->w5.dx * DXYZ_SCALE;
		tp->msg_data.reference.dy = ((m->w5.dy_h << 8) | m->w6.dy_l) * DXYZ_SCALE;
		tp->msg_data.reference.dz = m->w6.dz * DXYZ_SCALE;
	    } else 
		tp->msg_data.reference.sense = invalid;
	}
	break;
    case 5:
	for (n = 0; n < (unsigned)len; n++) {
	    struct consat_t *csp = &tp->msg_data.conhealth.sat[n];
	    struct b_health_t *m = &msg->msg_type.type5.health[n];

	    csp->ident = m->sat_id;
	    csp->iodl = m->issue_of_data_link!=0;
	    csp->health = m->data_health;
	    /*@i@*/csp->snr = (m->cn0?(m->cn0+CNR_OFFSET):SNR_BAD);
	    csp->health_en = m->health_enable;
	    csp->new_data = m->new_nav_data!=0;
	    csp->los_warning = m->loss_warn!=0;
	    csp->tou = m->time_unhealthy*TU_SCALE;
	}
	tp->msg_data.conhealth.nentries = n;
	break;
    case 7:
	for (w = 0; w < (unsigned)len; w++) {
	    struct station_t *np = &tp->msg_data.almanac.station[n];
	    struct b_station_t *mp = &msg->msg_type.type7.almanac[w];

	    np->latitude = mp->w3.lat * LA_SCALE;
	    /*@i@*/np->longitude = ((mp->w3.lon_h << 8) | mp->w4.lon_l) * LO_SCALE;
	    np->range = mp->w4.range;
	    np->frequency = (((mp->w4.freq_h << 6) | mp->w5.freq_l) * FREQ_SCALE) + FREQ_OFFSET;
	    np->health = mp->w5.health;
	    np->station_id = mp->w5.station_id,
	    np->bitrate = tx_speed[mp->w5.bit_rate];
	    n++;
	}
	tp->msg_data.almanac.nentries = (unsigned)(len/3);
	break;
    case 16:
	/*@ -boolops @*/
	for (w = 0; w < (unsigned)len; w++){
	    if (!msg->msg_type.type16.txt[w].byte1) {
		break;
	    }
	    tp->msg_data.message[n++] = (char)(msg->msg_type.type16.txt[w].byte1);
	    if (!msg->msg_type.type16.txt[w].byte2) {
		break;
	    }
	    tp->msg_data.message[n++] = (char)(msg->msg_type.type16.txt[w].byte2);
	    if (!msg->msg_type.type16.txt[w].byte3) {
		break;
	    }
	    tp->msg_data.message[n++] = (char)(msg->msg_type.type16.txt[w].byte3);
	}
	/*@ +boolops @*/
	tp->msg_data.message[n++] = '\0';
	break;

	case 18:
		tp->msg_data.phase = &(tp->msg_data.obs_data[0]);
		tp->msg_data.gnsstime = (long) msg->msg_type.type18.w3.gnsstime;
		tp->msg_data.phase->frequency = (uint) msg->msg_type.type18.w3.frequency;
		
		for (w = 0; w < (unsigned) ((len-1)/2); w++) {
			tp->msg_data.phase->obs[w].multiple = (bool) msg->msg_type.type18.phaseobs[w].w4.multiple;
			tp->msg_data.phase->obs[w].cacode = (bool) msg->msg_type.type18.phaseobs[w].w4.cacode;
			tp->msg_data.phase->obs[w].GPS = (bool) msg->msg_type.type18.phaseobs[w].w4.GPS;
			tp->msg_data.phase->obs[w].prn = (uint) msg->msg_type.type18.phaseobs[w].w4.prn;
			tp->msg_data.phase->obs[w].quality = (uint) msg->msg_type.type18.phaseobs[w].w4.quality;
			tp->msg_data.phase->obs[w].continuityloss = (uint) msg->msg_type.type18.phaseobs[w].w4.continuityloss;
			ltmp = (long) ((msg->msg_type.type18.phaseobs[w].w4.phase_h << 24) | msg->msg_type.type18.phaseobs[w].w5.phase_l);
			ltmp -= 1;
			ltmp = ~ltmp;
			tp->msg_data.phase->obs[w].value = (double)ltmp;
			tp->msg_data.phase->obs[w].value /= 256.0;

		}
		tp->msg_data.numobs = w;
		s = (tp->msg_data.phase->frequency == 0) ? RTCM_L1 : RTCM_L2;
		if (s != 0) {
			memcpy(&(tp->msg_data.obs_data[s]), tp->msg_data.phase, sizeof(tp->msg_data.obs_data[s]));
			tp->msg_data.phase = &(tp->msg_data.obs_data[s]);
		}
		tp->msg_data.phase->obs_avail = true;
		break;

	case 19:
		tp->msg_data.pseudo = &(tp->msg_data.obs_data[0]);
		tp->msg_data.gnsstime = (long) msg->msg_type.type19.w3.gnsstime;
		tp->msg_data.smth = (uint) msg->msg_type.type19.w3.smoothing;
		tp->msg_data.pseudo->frequency = (uint) msg->msg_type.type19.w3.frequency;
		
		for (w = 0; w < (unsigned) ((len-1)/2); w++) {
			tp->msg_data.pseudo->obs[w].multiple = (bool) msg->msg_type.type19.pseudoobs[w].w4.multiple;
			tp->msg_data.pseudo->obs[w].cacode = (bool) msg->msg_type.type19.pseudoobs[w].w4.cacode;
			tp->msg_data.pseudo->obs[w].GPS = (bool) msg->msg_type.type19.pseudoobs[w].w4.GPS;
			tp->msg_data.pseudo->obs[w].prn = (uint) msg->msg_type.type19.pseudoobs[w].w4.prn;
			tp->msg_data.pseudo->obs[w].quality = (uint) msg->msg_type.type19.pseudoobs[w].w4.quality;
			tp->msg_data.pseudo->obs[w].multipath = (uint) msg->msg_type.type19.pseudoobs[w].w4.multipath;
			ultmp = (long) ((msg->msg_type.type19.pseudoobs[w].w4.pseudo_h << 24) | msg->msg_type.type19.pseudoobs[w].w5.pseudo_l);
			tp->msg_data.pseudo->obs[w].value = (double)ultmp * 0.02;

		}
		tp->msg_data.numobs = w;

		s = RTCM_C1;
		if (tp->msg_data.phase->obs[0].cacode == 0) {
			s = (tp->msg_data.phase->frequency == 0) ? RTCM_P1 : RTCM_P2;
		}
		if (s != 0) {
			memcpy(&(tp->msg_data.obs_data[s]), tp->msg_data.pseudo, sizeof(tp->msg_data.obs_data[s]));
			tp->msg_data.pseudo = &(tp->msg_data.obs_data[s]);
		}
		tp->msg_data.pseudo->obs_avail = true;
		
		break;

	case 22:
		tp->msg_data.ecef_delta.l1_x_delta = msg->msg_type.type22.w3.l1_x_delta;
		tp->msg_data.ecef_delta.l1_y_delta = msg->msg_type.type22.w3.l1_y_delta;
		tp->msg_data.ecef_delta.l1_z_delta = msg->msg_type.type22.w3.l1_z_delta;
		tp->msg_data.ecef_delta.haveHI = 0;
		tp->msg_data.ecef_delta.haveL2 = 0;
		
		if (len >= 2) {
			tp->msg_data.ecef_delta.haveHI = 1;
			tp->msg_data.ecef_delta.GPS = (bool) msg->msg_type.type22.w4.GPS;
			tp->msg_data.ecef_delta.at = (bool) msg->msg_type.type22.w4.at;
			tp->msg_data.ecef_delta.ar = (bool) msg->msg_type.type22.w4.ar;
			tp->msg_data.ecef_delta.nh = (bool) msg->msg_type.type22.w4.nh;
			tp->msg_data.ecef_delta.l1_phase_height = msg->msg_type.type22.w4.HI;
		}
		if (len >= 3) {
			tp->msg_data.ecef_delta.haveL2 = 1;
			tp->msg_data.ecef_delta.l2_x_delta = msg->msg_type.type22.w5.l2_x_delta;
			tp->msg_data.ecef_delta.l2_y_delta = msg->msg_type.type22.w5.l2_y_delta;
			tp->msg_data.ecef_delta.l2_z_delta = msg->msg_type.type22.w5.l2_z_delta;
		}
		break;

	case 23:
		tp->msg_data.antenna.reference = (bool)msg->msg_type.type23.w3.ar;
		tp->msg_data.antenna.serial = (bool)msg->msg_type.type23.w3.sf;
		tp->msg_data.antenna.len_desc = msg->msg_type.type23.w3.nad;
		int nad = msg->msg_type.type23.w3.nad;
		if (nad >= 1)
			tp->msg_data.antenna.descriptor[0] = (char)msg->msg_type.type23.w3.byte1;
		if (nad >= 2)
			tp->msg_data.antenna.descriptor[1] = (char)msg->msg_type.type23.w3.byte2;
		
		int pos;
		int serial = -1;
		for (pos = 0 ; pos < ((len-1)*3); pos++) {
			int index = pos/3;
			int byte = pos%3;
			unsigned char val = 0;

			switch (byte) {
				case 0:
					val = (unsigned char)msg->msg_type.type23.info[index].byte1;
					break;
				case 1:
					val = (unsigned char)msg->msg_type.type23.info[index].byte2;
					break;
				case 2:
					val = (unsigned char)msg->msg_type.type23.info[index].byte3;
					break;
			}
			if (pos < (nad - 2)) {
				tp->msg_data.antenna.descriptor[pos+2] = (char) val;
				continue;
			}
			if (pos == (nad - 2)) {
				tp->msg_data.antenna.descriptor[pos+2] = '\0';
				tp->msg_data.antenna.setupid = (int) val;
				continue;
			}
			if (pos == (nad - 1)) {
				tp->msg_data.antenna.len_serial = (int) (val & 0x1f);
				serial = 0;
				continue;
			}
			if (serial < tp->msg_data.antenna.len_serial) {
				tp->msg_data.antenna.serialnum[serial] = (char) val;
			} else { // done withreading record
				tp->msg_data.antenna.serialnum[serial] = '\0';
				break;
			}
		}
		break;

	case 24:
		memset(&u38tmp,0,sizeof(u38tmp));
		u38tmp |= msg->msg_type.type24.w3.ecef_x_h;
		u38tmp <<= 14;
		u38tmp |= msg->msg_type.type24.w4.ecef_x_l;
		if (u38tmp & 0x2000000000LL) {			// negative value
			u38tmp -= 1; u38tmp = ~u38tmp;		// 2's complement
			u38tmp &= 0x3fffffffffLL; 			// set bits above 38 to 0
			tp->msg_data.coord.ecef_x = sprintf(ctmp,"-%lld",u38tmp); // set to -value 
		} else {								// positive value
			tp->msg_data.coord.ecef_x = sprintf(ctmp,"%lld",u38tmp);
		}
		tp->msg_data.coord.ecef_x = strtod(ctmp,NULL);
		tp->msg_data.coord.ecef_x *= .0001;
		
		memset(&u38tmp,0,sizeof(u38tmp));
		u38tmp |= msg->msg_type.type24.w4.ecef_y_h;
		u38tmp <<= 24;
		u38tmp |= msg->msg_type.type24.w5.ecef_y_m;
		u38tmp <<= 6;
		u38tmp |= msg->msg_type.type24.w6.ecef_y_l;
		if (u38tmp & 0x2000000000LL) {			// negative value
			u38tmp -= 1; u38tmp = ~u38tmp;		// 2's complement
			u38tmp &= 0x3fffffffffLL; 			// set bits above 38 to 0
			tp->msg_data.coord.ecef_y = sprintf(ctmp,"-%lld",u38tmp); // set to -value 
		} else {								// positive value
			tp->msg_data.coord.ecef_y = sprintf(ctmp,"%lld",u38tmp);
		}
		tp->msg_data.coord.ecef_y = strtod(ctmp,NULL);
		tp->msg_data.coord.ecef_y *= .0001;

		memset(&u38tmp,0,sizeof(u38tmp));
		u38tmp |= msg->msg_type.type24.w6.ecef_z_h;
		u38tmp <<= 22;
		u38tmp |= msg->msg_type.type24.w7.ecef_z_l;
		if (u38tmp & 0x2000000000LL) {			// negative value
			u38tmp -= 1; u38tmp = ~u38tmp;		// 2's complement
			u38tmp &= 0x3fffffffffLL; 			// set bits above 38 to 0
			tp->msg_data.coord.ecef_z = sprintf(ctmp,"-%lld",u38tmp); // set to -value 
		} else {								// positive value
			tp->msg_data.coord.ecef_z = sprintf(ctmp,"%lld",u38tmp);
		}
		tp->msg_data.coord.ecef_z = strtod(ctmp,NULL);
		tp->msg_data.coord.ecef_z *= .0001;

		tp->msg_data.coord.haveHI = (bool) msg->msg_type.type24.w7.haveHI;
		tp->msg_data.coord.GPS = (bool) msg->msg_type.type24.w7.GPS;
		tp->msg_data.coord.HI = (double) (msg->msg_type.type24.w8.HI / 1e4);

		break;

    default:
	memcpy(tp->msg_data.words, msg->msg_type.rtcm_msgunk, (RTCM_WORDS_MAX-2)*sizeof(isgps30bits_t));
	break;
    }
}

uint32_t CRC24(long size, const unsigned char *buf)
{
    uint32_t crc = 0;
    int i;

    while(size--)
    {
        crc ^= (*buf++) << (16);
        for(i = 0; i < 8; i++)
        {
            crc <<= 1;
            if(crc & 0x1000000)
                crc ^= 0x01864cfb;
        }
    }
    return crc;
}

long long int getbitsMSB_signed(unsigned char *src, unsigned int bitsize)
{
	unsigned long long ret = getbitsMSB(src, bitsize);
	return((long long int)(((long long int )(ret<<(64 - bitsize)))>>(64 - bitsize)));
}

unsigned long long getbitsMSB(unsigned char *src, unsigned int bitsize )
{
    
    static unsigned char current_byte = 0;
    static unsigned char output_mask = 0x80;
    int i;
    unsigned long long dest = 0;
    if (src == NULL) {
        output_mask = 0x80;
        current_byte = 0;
        return(0);
    }
    for (i = bitsize-1; i >=0 ; i-- ) {
        if ( src[current_byte] & output_mask )
            dest += 1;
        if (i > 0) dest <<= 1;
        output_mask >>= 1;
        if ( output_mask == 0 ) {
            output_mask = 0x80;
            ++current_byte;
        }
    }
    return(dest);
}

int putbitsMSB_signed(unsigned char *dest, long long src, unsigned int bitsize )
{
	unsigned long long ul = ((unsigned long long )(src<<(64 - bitsize)))>>(64 - bitsize);
	int i = putbitsMSB(dest, (unsigned long long)ul, bitsize);
	return(i);
}

int putbitsMSB(unsigned char *dest, unsigned long long src, unsigned int bitsize )
{
    
    static unsigned char current_byte = 0;
    static unsigned char output_mask = 0x80;
    if (dest == NULL) {
        output_mask = 0x80;
        current_byte = 0;
        return(0);
    }
    int i;
    for (i = bitsize-1; i >=0 ; i-- ) {
        if ( src & (long long)(0x1LL<<(long long)i) )
            dest[current_byte] |= output_mask;

        output_mask >>= 1;
        if ( output_mask == 0 ) {
            output_mask = 0x80;
            dest[++current_byte] = 0;
        }
    }
    return(current_byte);
}


void rtcm3_unpack(/*@out@*/struct rtcm_t *tp, unsigned char *buf)
{
    int i;
    long long int i38tmp;
    unsigned long msg_len = 0;
    unsigned char c;
    unsigned short s;

    getbitsMSB(NULL, 0);
    c = (char) getbitsMSB(buf, 8);
    c = (char) getbitsMSB(buf, 6);
    msg_len = (short) getbitsMSB(buf, 10);

    tp->type = (short) getbitsMSB(buf, 12);
    tp->refstaid = (short) getbitsMSB(buf, 12);                 // reference station ID

    switch (tp->type) {
        case 1001: case 1002: case 1003: case 1004:
 
            tp->msg_data.gpstow    = (unsigned long) getbitsMSB(buf, 30);  // time of week in ms
            tp->msg_data.gnss_sync = (bool) getbitsMSB(buf, 1);                             // time of week in ms
            tp->msg_data.numobs    = (char) getbitsMSB(buf, 5);    // Number of satellite obs
            tp->msg_data.smth      = (bool) getbitsMSB(buf, 1);
            tp->msg_data.smth_int  = (char) getbitsMSB(buf, 3);                             // Frequency of data smoothing
            for (i = 0; i < tp->msg_data.numobs; i++) {
                char prn = (char) getbitsMSB(buf, 6);  // Sat ID
		bool cacode = (bool) getbitsMSB(buf, 1);  // L1 indicator

		int c1p1 = (cacode == 0) ? RTCM_C1 : RTCM_P1;

                tp->msg_data.obs_data[RTCM_L1].obs[i].cacode = cacode;
                tp->msg_data.obs_data[RTCM_L1].obs[i].prn = prn;

                tp->msg_data.obs_data[c1p1].obs[i].cacode = cacode;
                tp->msg_data.obs_data[c1p1].obs[i].prn = prn;

                unsigned long long pr  = getbitsMSB(buf, 24);  // L1 pseudorange
		double c1p1_pr = (double)pr * 0.02;

                long long ph = getbitsMSB_signed(buf, 20);                // L1 phase - pseudo
		double l1_pr = (double) ph * 0.0005;

                tp->msg_data.obs_data[RTCM_L1].obs[i].locktime = getbitsMSB(buf, 7);                // L1 locktime
		tp->msg_data.obs_data[RTCM_L1].obs_avail = true;
		tp->msg_data.obs_data[c1p1].obs_avail = true;

                if ((tp->type == 1002) || (tp->type == 1004)) {
                    int pma = getbitsMSB(buf, 8);              // L1 pseudo modulus
                    c1p1_pr += (double) pma * LIGHTSPEED_M;
                    tp->msg_data.obs_data[RTCM_L1].obs[i].quality  = getbitsMSB(buf, 8);        // L1 CNR
                }
                tp->msg_data.obs_data[c1p1].obs[i].value = c1p1_pr;
                tp->msg_data.obs_data[RTCM_L1].obs[i].value = ((c1p1_pr + l1_pr) / WV_L1);

                if ((tp->type == 1003) || (tp->type == 1004)) {
                    tp->msg_data.obs_data[RTCM_L2].obs[i].cacode = getbitsMSB(buf, 2);          // L2 indicator
		    int s = getbitsMSB_signed(buf, 14);  					// L2 - L1 pseudo
                    tp->msg_data.obs_data[RTCM_P2].obs[i].value = ((double)s * 0.02) + c1p1_pr;

                    ph = getbitsMSB_signed(buf, 20);						// L2 phase - L1 pseudo
		    double l2_pr = (double) ph * 0.0005;
                    tp->msg_data.obs_data[RTCM_L2].obs[i].value = ((c1p1_pr + l2_pr) / WV_L2);

                    tp->msg_data.obs_data[RTCM_L2].obs[i].locktime = getbitsMSB(buf, 7);        // L2 locktime
                    tp->msg_data.obs_data[RTCM_L2].obs[i].prn = prn;
                    tp->msg_data.obs_data[RTCM_P2].obs[i].prn = prn;
		    tp->msg_data.obs_data[RTCM_L2].obs_avail = true;
		    tp->msg_data.obs_data[RTCM_P2].obs_avail = true;
                }
                if (tp->type == 1004) {
                    tp->msg_data.obs_data[RTCM_L2].obs[i].quality = getbitsMSB(buf, 8);                  // L2 CNR
                }
            }

            break;
        case 1005:
        case 1006:
            getbitsMSB(buf, 6);                          // reserved
            tp->msg_data.coord.GPS = getbitsMSB(buf, 1);                          // GPS
            getbitsMSB(buf, 1);                          // not glonass
            getbitsMSB(buf, 1);                          // not galileo
            getbitsMSB(buf, 1);                          // physical reference station

            i38tmp = getbitsMSB(buf, 38);                    // ECEF_X
            tp->msg_data.coord.ecef_x = ((double)i38tmp * .0001);

            getbitsMSB(buf, 1);                          // SRO
            getbitsMSB(buf, 1);                          // reserved

            i38tmp = getbitsMSB(buf, 38);                    // ECEF_X
            tp->msg_data.coord.ecef_y = ((double)i38tmp * .0001);

            getbitsMSB(buf, 2);                          // reserved

            i38tmp = getbitsMSB(buf, 38);                    // ECEF_X
            tp->msg_data.coord.ecef_z = ((double)i38tmp * .0001);

            if (tp->type == 1006) {
                s = getbitsMSB(buf, 16);
                tp->msg_data.coord.HI = ((double) s * .0001); // Antenna Height
            }

            break;
        case 1007:
        case 1008:

            s = getbitsMSB(buf, 8);
            for (i = 0; i < s; i++ )
                tp->msg_data.antenna.descriptor[i] = getbitsMSB(buf, 8);

            tp->msg_data.antenna.setupid = getbitsMSB(buf, 8);

            if (tp->type == 1008) {
                s = getbitsMSB(buf, 8);
                for (i = 0; i < s; i++ )
                    tp->msg_data.antenna.serialnum[i] = getbitsMSB(buf, 8);
            }
            break;
        default:
            getbitsMSB(NULL,0);
            return;
    }

    return; // return total bytes

}

int rtcm3_repack(struct rtcm_t *tp, unsigned char *buf)
/* repack the content fields into the raw bits */
{
    int i;
    long long int i38tmp;
    unsigned long msg_len = 0;
    unsigned short s;

    buf[0] = 0;
    putbitsMSB(NULL,0,0);
    putbitsMSB(buf, 0xD3, 8);
    putbitsMSB(buf, 0x0, 6);

    switch (tp->type) {
        case 1001: case 1002: case 1003: case 1004:
            msg_len = 8;
            int numobs = tp->msg_data.numobs;
            double obs_size = 0.0;
            if (tp->type == 1001)
                obs_size = 7.25 * numobs;
            if (tp->type == 1002)
                obs_size = 9.25 * numobs;
            if (tp->type == 1003)
                obs_size = 12.625 * numobs;
            if (tp->type == 1004)
                obs_size = 15.625 * numobs;

            obs_size = ceil(obs_size + 0.001);
            msg_len += (int) obs_size;
 
            putbitsMSB(buf, msg_len, 10); // message length w/o header/checksum

            putbitsMSB(buf, tp->type, 12);                     // message type
            putbitsMSB(buf, tp->refstaid, 12);                 // reference station ID
            putbitsMSB(buf, tp->msg_data.gpstow, 30);  // time of week in ms
            putbitsMSB(buf, 0, 1);                             // time of week in ms
            putbitsMSB(buf, numobs, 5);                        // Number of satellite obs
            int smooth = (tp->msg_data.smth) ? 1 : 0;
            putbitsMSB(buf, smooth, 1);                        // Is the data smoothed?
            putbitsMSB(buf, 0, 3);                             // Frequency of data smoothing
            unsigned int prn_found = 0;
            for (i = 0; i < numobs; i++) {
                unsigned int prn = tp->msg_data.obs_data[RTCM_L1].obs[i].prn;
		prn_found &= (1 << (prn-1));
                putbitsMSB(buf, prn, 6);  // Sat ID

		int c1p1;
		if (tp->msg_data.obs_data[RTCM_P1].obs[i].usable) {
			c1p1 = RTCM_P1;
		} else if (tp->msg_data.obs_data[RTCM_C1].obs[i].usable) {
			c1p1 = RTCM_C1;
		} else { // Set to C1 so other values can be recreated later
			c1p1 = RTCM_C1;
		}

                putbitsMSB(buf, tp->msg_data.obs_data[c1p1].obs[i].cacode, 1);  // L1 indicator
		double ps = tp->msg_data.obs_data[c1p1].obs[i].value;

		int amb = (int) (ps / LIGHTSPEED_M);

		double ps_mod = ((ps - ((double)amb * LIGHTSPEED_M)) / 0.02);
		unsigned long long d = (unsigned long long) llround(ps_mod);

		double c1p1_ps = (((double) d * 0.02) +  ((double)amb * LIGHTSPEED_M));

                putbitsMSB(buf, (unsigned long long) d, 24);  // L1 pseudorange

		// current phase from l1 value from receiver
		double l1c = tp->msg_data.obs_data[RTCM_L1].obs[i].value;

		// current integer cycles difference between l1 phase and l1 pseudorange as stored in P1/C1
		if ((! phase_ambiguity[prn].l1_valid) || (tp->msg_data.obs_data[RTCM_L1].obs[i].continuityloss)) {
			phase_ambiguity[prn].l1_valid = 1;
			phase_ambiguity[prn].l1_start = tp->msg_data.gpstow - 1000; // set to previous second.
			phase_ambiguity[prn].l1_cycles = lround(((double)c1p1_ps / (double)WV_L1) - (double)l1c);

		}

#define NEWPHASECALC
#ifdef NEWPHASECALC
		// Actual L1 phase range within 1 cycle of p1
		long l1_diff = lround(((double)c1p1_ps / (double)WV_L1) - (double)l1c);
		// Don't alter the reading if it already in range of the output
		if (labs(l1_diff) < (long)(262.1435 / WV_L1))
			l1_diff = 0;
		double ph = ((l1c + l1_diff) * WV_L1);

#else
		// Actual L1 phase range
		double ph = ((l1c + phase_ambiguity[prn].l1_cycles) * WV_L1);

		// account for possible rollover
		if (fabs(ph - c1p1_ps) > 262.1435) {
			if ((ph - c1p1_ps) > 262.1435)
				phase_ambiguity[prn].l1_cycles -= 1500;
			else
				phase_ambiguity[prn].l1_cycles += 1500;
			ph = ((l1c + (double)phase_ambiguity[prn].l1_cycles) * WV_L1);
		}
#endif

		int j;
		if (tp->msg_data.obs_data[RTCM_L1].obs[i].usable) {
			j = (int)((ph - c1p1_ps) / 0.0005);
		} else {
			j = 1 << 19;
			phase_ambiguity[prn].l1_valid = 0;
			phase_ambiguity[prn].l1_start = 0;
			phase_ambiguity[prn].l1_cycles = 0;
		}

                putbitsMSB_signed(buf, j, 20);                 // L1 phase - pseudo

                unsigned int secs = (unsigned int) ((tp->msg_data.gpstow - phase_ambiguity[prn].l1_start) / 1000.0);
		if (secs < 24) {
			;	
		} else if (secs < 72) {
			secs = (secs + 24)/2;
		} else if (secs < 168) {
			secs = (secs + 120)/4;
		} else if (secs < 360) {
			secs = (secs + 408)/8;
		} else if (secs < 744) {
			secs = (secs + 1176)/16;
		} else if (secs < 937) {
			secs = (secs + 3096)/32;
		} else {
			secs = 127;
		}
	
                tp->msg_data.obs_data[RTCM_L1].obs[i].locktime = secs;

                putbitsMSB(buf, tp->msg_data.obs_data[RTCM_L1].obs[i].locktime, 7);                  // L1 locktime
                if ((tp->type == 1002) || (tp->type == 1004)) {
                    putbitsMSB(buf, amb, 8);              // L1 pseudo modulus
                    putbitsMSB(buf, tp->msg_data.obs_data[RTCM_L1].obs[i].quality, 8);              // L1 CNR
                }
                if ((tp->type == 1003) || (tp->type == 1004)) {

                    if (tp->msg_data.obs_data[RTCM_P2].obs[i].usable) {
//			j = (int)(((tp->msg_data.obs_data[RTCM_P2].obs[i].value - ((double)amb * LIGHTSPEED_M)) - c1p1_ps) / 0.02);
//			j = (int) round((tp->msg_data.obs_data[RTCM_P2].obs[i].value - tp->msg_data.obs_data[c1p1].obs[i].value)/0.02);
			j = (int) round((tp->msg_data.obs_data[RTCM_P2].obs[i].value - c1p1_ps)/0.02);
			putbitsMSB(buf, tp->msg_data.obs_data[RTCM_P2].obs[i].cacode, 2);                  // L2 indicator
                    } else {
			j = 1 << 13;
                        putbitsMSB(buf, 1, 2);                  // L2 indicator hardwired to 1
		    }
                    putbitsMSB_signed(buf, j, 14);                 // L2 - L1 pseudo

	    	double l2c = tp->msg_data.obs_data[RTCM_L2].obs[i].value;

	    	if ((! phase_ambiguity[prn].l2_valid) || (tp->msg_data.obs_data[RTCM_L2].obs[i].continuityloss)) {
			phase_ambiguity[prn].l2_valid = 1;
			phase_ambiguity[prn].l2_start = tp->msg_data.gpstow - 1000; // set to previous second.
			phase_ambiguity[prn].l2_cycles = lround(((double)c1p1_ps / (double)WV_L2) - (double)l2c);
	    	}

#ifdef NEWPHASECALC
		// Actual L2 phase range within 1 cycle of p1
		long l2_diff = lround(((double)c1p1_ps / (double)WV_L2) - (double)l2c);
		// Don't alter the reading if it already in range of the output
		if (labs(l2_diff) < (long)(262.1435 / WV_L2))
			l2_diff = 0;
                double ph_l2 = ((l2c + l2_diff) * WV_L2);

#else
		double ph_l2 = ((l2c + phase_ambiguity[prn].l2_cycles) * WV_L2);
		if (fabs(ph_l2 - c1p1_ps) > 262.1435) {
			if ((ph_l2 - c1p1_ps) > 262.1435)
				phase_ambiguity[prn].l2_cycles -= 1500;
			else
				phase_ambiguity[prn].l2_cycles += 1500;
			ph_l2 = ((l2c + (double)phase_ambiguity[prn].l2_cycles) * WV_L2);
		}
#endif

		if (tp->msg_data.obs_data[RTCM_L2].obs[i].usable) {
			j = (int)((ph_l2 - c1p1_ps) / 0.0005);
		} else {
			j = 1 << 19;
			phase_ambiguity[prn].l2_valid = 0;
			phase_ambiguity[prn].l2_start = 0;
			phase_ambiguity[prn].l2_cycles = 0;
		}
                    putbitsMSB_signed(buf, j, 20);                 // L2 phase - L1 pseudo

		    secs = (unsigned int) ((tp->msg_data.gpstow - phase_ambiguity[prn].l2_start) / 1000.0);
                    if (secs < 24) {
                        ;
                    } else if (secs < 72) {
                        secs = (secs + 24)/2;
                    } else if (secs < 168) {
                        secs = (secs + 120)/4;
                    } else if (secs < 360) {
                        secs = (secs + 408)/8;
                    } else if (secs < 744) {
                        secs = (secs + 1176)/16;
                    } else if (secs < 937) {
                        secs = (secs + 3096)/32;
                    } else {
                        secs = 127;
                    } 

                    tp->msg_data.obs_data[RTCM_L2].obs[i].locktime = secs;

                    putbitsMSB(buf, tp->msg_data.obs_data[RTCM_L2].obs[i].locktime, 7);                  // L2 locktime
                }
                if (tp->type == 1004) {
                    putbitsMSB(buf, tp->msg_data.obs_data[RTCM_L2].obs[i].quality, 8);                  // L2 CNR
                }
            }
            // Reset ambiguity parameters for prn's not found in current record (this may be a bad hack)
            for (i = 1; i <= 32;  i++) {
		if (prn_found && (1 << (i - 1)))
			memset(&phase_ambiguity[i],0,sizeof(struct ambiguity_t));
            }
            for (i = putbitsMSB(buf,0,0); i <= msg_len+3; i++) {
                i = putbitsMSB(buf,0,1);
            }

            break;
        case 1005:
        case 1006:
            msg_len = 19;
            if (tp->type == 1006)
                msg_len += 2;

            putbitsMSB(buf, msg_len, 10);                   // message length w/o header/checksum

            putbitsMSB(buf, tp->type, 12);                  // message type
            putbitsMSB(buf, tp->refstaid, 12);              // reference station ID
            putbitsMSB(buf, 0, 6);                          // reserved
            putbitsMSB(buf, 1, 1);                          // GPS
            putbitsMSB(buf, 0, 1);                          // not glonass
            putbitsMSB(buf, 0, 1);                          // not galileo
            putbitsMSB(buf, 0, 1);                          // physical reference station

            memset(&i38tmp,0,sizeof(i38tmp));
            i38tmp = (long long int)(tp->msg_data.coord.ecef_x/.0001);
            i38tmp &= 0x3fffffffffLL;                       // set bits above 38 to 0
            putbitsMSB(buf, i38tmp, 38);                    // ECEF_X

            putbitsMSB(buf, 0, 1);                          // SRO
            putbitsMSB(buf, 0, 1);                          // reserved

            memset(&i38tmp,0,sizeof(i38tmp));
            i38tmp = (long long int)(tp->msg_data.coord.ecef_y/.0001);
            i38tmp &= 0x3fffffffffLL;                       // set bits above 38 to 0
            putbitsMSB(buf, i38tmp, 38);                    // ECEF_Y

            putbitsMSB(buf, 0, 2);                          // reserved

            memset(&i38tmp,0,sizeof(i38tmp));
            i38tmp = (long long int)(tp->msg_data.coord.ecef_z/.0001);
            i38tmp &= 0x3fffffffffLL;                       // set bits above 38 to 0
            putbitsMSB(buf, i38tmp, 38);                    // ECEF_Z

            if (tp->type == 1006) {
		s = 0;
		s = (unsigned short) (tp->msg_data.coord.HI / .0001);
                putbitsMSB(buf, s, 16); // Antenna Height
	    }

            break;
        case 1007:
        case 1008:

            msg_len = 5;
            int desc = 0;
            int snum = 0;
            desc = strlen(tp->msg_data.antenna.descriptor);
            msg_len += desc;

            if (tp->type == 1008) {
                snum = (int)strlen(tp->msg_data.antenna.serialnum);
                msg_len += snum + 1;
            }

            putbitsMSB(buf, msg_len, 10); // message length w/o header/checksum
            putbitsMSB(buf, tp->type, 12);      // message type
            putbitsMSB(buf, tp->refstaid, 12);  // reference station ID

            putbitsMSB(buf, desc, 8);
            for (i = 0; i < desc; i++ )
                putbitsMSB(buf, tp->msg_data.antenna.descriptor[i], 8);

            putbitsMSB(buf, (long long int)tp->msg_data.antenna.setupid, 8);

            if (tp->type == 1008) {
                putbitsMSB(buf, snum, 8);
                for (i = 0; i < snum; i++ )
                    putbitsMSB(buf, (long long int)tp->msg_data.antenna.serialnum[i], 8);
            }
            break;
        default:
            putbitsMSB(NULL,0,0);
            return(0);
    }

    unsigned long crc;
    crc = CRC24(msg_len + 3, buf); // add header to msg_len
    putbitsMSB(buf, crc, 24);

    return(msg_len + 6); // return total bytes

}



bool rtcm_repack(struct rtcm_t *tp, isgps30bits_t *buf)
/* repack the content fields into the raw bits */
{
    int len, sval;
    unsigned int n, w, uval;
    struct rtcm_msg_t  *msg = (struct rtcm_msg_t *)buf;
    struct rtcm_msghw1 *wp  = (struct rtcm_msghw1 *)buf;

    unsigned int itmp;
    unsigned long long int u38tmp;
    long ltmp;
    double dtmp;
    char ctmp[BUFSIZ];

    msg->w1.preamble = PREAMBLE_PATTERN;

    msg->w1.msgtype = tp->type;
    msg->w2.frmlen = tp->length;
    msg->w2.zcnt = (unsigned) round(tp->zcount / ZCOUNT_SCALE);
    msg->w1.refstaid = tp->refstaid;
    msg->w2.sqnum = tp->seqnum;
    msg->w2.stathlth = tp->stathlth;

    len = (int)tp->length;
    n = 0;
    switch (tp->type) {
    case 1:	/* S */
    case 9:
	{
	    struct b_correction_t    *m = &msg->msg_type.type1.corrections[0];

	    while (len >= 0) {
		if (len >= 2) {
		    struct rangesat_t *ssp = &tp->msg_data.ranges.sat[n];
		    m->w3.satident1 = ssp->ident;
		    m->w3.udre1 = ssp->udre;
		    m->w4.issuedata1 = ssp->issuedata;
		    m->w3.scale1 = (unsigned)((ssp->rangerr > MAXPCSMALL) ||
					      (ssp->rangerr < (-MAXPCSMALL)) ||
					      (ssp->rangerate > MAXRRSMALL) ||
					      (ssp->rangerate < (-MAXRRSMALL)));
		    m->w3.pc1 = (int) round(ssp->rangerr / (m->w3.scale1 ? PCLARGE : PCSMALL));
		    m->w4.rangerate1 = (int) round(ssp->rangerate / (m->w3.scale1 ? RRLARGE : RRSMALL));
		    n++;
		}
		if (len >= 4) {
		    struct rangesat_t *ssp = &tp->msg_data.ranges.sat[n];
		    m->w4.satident2 = ssp->ident;
		    m->w4.udre2 = ssp->udre;
		    m->w6.issuedata2 = ssp->issuedata;
		    m->w4.scale2 = (unsigned)((ssp->rangerr > MAXPCSMALL) ||
					      (ssp->rangerr < (-MAXPCSMALL)) ||
					      (ssp->rangerate > MAXRRSMALL) ||
					      (ssp->rangerate < (-MAXRRSMALL)));
		    m->w5.pc2 = (int) round(ssp->rangerr / (m->w4.scale2 ? PCLARGE : PCSMALL));
		    m->w5.rangerate2 = (int) round(ssp->rangerate / (m->w4.scale2 ? RRLARGE : RRSMALL));
		    n++;
		}
		if (len >= 5) {
		    struct rangesat_t *ssp = &tp->msg_data.ranges.sat[n];
		    m->w6.satident3 = ssp->ident;
		    m->w6.udre3 = ssp->udre;
		    m->w7.issuedata3 = ssp->issuedata;
		    m->w6.scale3 = (unsigned)((ssp->rangerr > MAXPCSMALL) ||
					      (ssp->rangerr < (-MAXPCSMALL)) ||
					      (ssp->rangerate > MAXRRSMALL) ||
					      (ssp->rangerate < (-MAXRRSMALL)));
		    sval = (int) round(ssp->rangerr / (m->w6.scale3 ? PCLARGE : PCSMALL));
		    /*@ -shiftimplementation @*/
		    m->w6.pc3_h = sval >> 8;
		    /*@ +shiftimplementation @*/
		    m->w7.pc3_l = (unsigned)sval & 0xff;
		    m->w7.rangerate3 = (int) round(ssp->rangerate / (m->w6.scale3 ? RRLARGE : RRSMALL));
		    n++;
		}
		len -= 5;
		m++;
	    }
	    tp->msg_data.ranges.nentries = n;
	}
	break;
    case 3:	/* R */
	if (tp->msg_data.ecef.valid) {
	    struct rtcm_msg3    *m = &msg->msg_type.type3;
	    unsigned x = (unsigned) round(tp->msg_data.ecef.x / XYZ_SCALE);
	    unsigned y = (unsigned) round(tp->msg_data.ecef.y / XYZ_SCALE);
	    unsigned z = (unsigned) round(tp->msg_data.ecef.z / XYZ_SCALE);

	    m->w4.x_l = x & 0xff;
	    m->w3.x_h = x >> 8;
	    m->w5.y_l = y & 0xffff;
	    m->w4.y_h = y >> 16;
	    m->w6.z_l = z & 0xffffff;
	    m->w5.z_h = z >> 24;
	}
	break;
    case 4:	/* D */
	if (tp->msg_data.reference.valid) {
	    struct rtcm_msg4    *m = &msg->msg_type.type4;

	    m->w3.dgnss = tp->msg_data.reference.system;
	    m->w3.dat = (unsigned)(tp->msg_data.reference.sense == global);
	    /*@ -predboolothers -type @*/
	    if (tp->msg_data.reference.datum[0])
		m->w3.datum_alpha_char1 = tp->msg_data.reference.datum[0];
	    else
		m->w3.datum_alpha_char1 = 0;
	    if (tp->msg_data.reference.datum[1])
		m->w3.datum_alpha_char2 = tp->msg_data.reference.datum[1];
	    else
		m->w3.datum_alpha_char2 = 0;
	    if (tp->msg_data.reference.datum[2])
		m->w4.datum_sub_div_char1 = tp->msg_data.reference.datum[2];
	    else
		m->w4.datum_sub_div_char1 = 0;
	    if (tp->msg_data.reference.datum[3])
		m->w4.datum_sub_div_char2 = tp->msg_data.reference.datum[3];
	    else
		m->w4.datum_sub_div_char2 = 0;
	    if (tp->msg_data.reference.datum[4])
		m->w4.datum_sub_div_char3 = tp->msg_data.reference.datum[4];
	    else
		m->w4.datum_sub_div_char3 = 0;
	    /*@ +predboolothers +type @*/
	    if (tp->msg_data.reference.system != unknown) {
		m->w5.dx = (uint)round(tp->msg_data.reference.dx / DXYZ_SCALE);
		uval = (uint)round(tp->msg_data.reference.dy / DXYZ_SCALE);
		m->w5.dy_h = uval >> 8;
		m->w6.dy_l = uval & 0xff;
		m->w6.dz = (uint)round(tp->msg_data.reference.dz / DXYZ_SCALE);
	    }
	}
	break;
    case 5:	/* C */
	for (n = 0; n < (unsigned)len; n++) {
	    struct consat_t *csp = &tp->msg_data.conhealth.sat[n];
	    struct b_health_t *m = &msg->msg_type.type5.health[n];

	    m->sat_id = csp->ident;
	    m->issue_of_data_link = (unsigned)csp->iodl;
	    m->data_health = csp->health;
	    m->cn0 = (csp->snr == SNR_BAD) ? 0 : (unsigned)csp->snr-CNR_OFFSET;
	    m->health_enable = csp->health_en;
	    m->new_nav_data = (unsigned)csp->new_data;
	    m->loss_warn = (unsigned)csp->los_warning;
	    m->time_unhealthy = (unsigned)(csp->tou / TU_SCALE);
	}
	break;
    case 7:	/* A */
	for (w = 0; w < (RTCM_WORDS_MAX - 2)/ 3; w++) {
	    struct station_t *np = &tp->msg_data.almanac.station[n++];
	    struct b_station_t *mp = &msg->msg_type.type7.almanac[w];

	    mp->w3.lat = (int) round(np->latitude / LA_SCALE);
	    sval = (int) round(np->longitude / LO_SCALE);
	    /*@ -shiftimplementation @*/
	    mp->w3.lon_h = sval >> 8;
	    /*@ +shiftimplementation @*/
	    mp->w4.lon_l = (unsigned)sval & 0xff;
	    mp->w4.range = np->range;
	    uval = (unsigned) round(((np->frequency-FREQ_OFFSET) / FREQ_SCALE));
	    mp->w4.freq_h = uval >> 6;
	    mp->w5.freq_l = uval & 0x3f;
	    mp->w5.health = np->health;
	    mp->w5.station_id = np->station_id;
	    mp->w5.bit_rate = 0;
	    for (uval = 0; uval < (unsigned)(sizeof(tx_speed)/sizeof(tx_speed[0])); uval++)
		if (tx_speed[uval] == np->bitrate) {
		    mp->w5.bit_rate = uval;
		    break;
		}
	    if (mp->w5.bit_rate == 0)
		return false;
	}
	tp->msg_data.almanac.nentries = n;
	break;
    case 16:	/* T */
	/*@ -boolops @*/
	for (w = 0; w < RTCM_WORDS_MAX - 2; w++){
	    if (!tp->msg_data.message[n]) {
		break;
	    }
	    msg->msg_type.type16.txt[w].byte1 = (unsigned)tp->msg_data.message[n++];
	    if (!tp->msg_data.message[n]) {
		break;
	    }
	    msg->msg_type.type16.txt[w].byte2 = (unsigned)tp->msg_data.message[n++];
	    if (!tp->msg_data.message[n]) {
		break;
	    }
	    msg->msg_type.type16.txt[w].byte3 = (unsigned)tp->msg_data.message[n++];
	}
	msg->w2.frmlen = w+1;
	/*@ +boolops @*/
	break;

	case 18:
		if (tp->msg_data.phase == NULL) break;
		msg->msg_type.type18.w3.gnsstime = (unsigned)tp->msg_data.gnsstime;
		msg->msg_type.type18.w3.frequency = (unsigned)tp->msg_data.phase->frequency;
		for (w = 0; w < tp->msg_data.numobs; w++) {
			msg->msg_type.type18.phaseobs[w].w4.multiple = (unsigned)tp->msg_data.phase->obs[w].multiple;
			msg->msg_type.type18.phaseobs[w].w4.cacode = (unsigned)tp->msg_data.phase->obs[w].cacode;
			msg->msg_type.type18.phaseobs[w].w4.GPS = (unsigned)tp->msg_data.phase->obs[w].GPS;
			msg->msg_type.type18.phaseobs[w].w4.prn = (unsigned)tp->msg_data.phase->obs[w].prn;
			msg->msg_type.type18.phaseobs[w].w4.quality = (unsigned)tp->msg_data.phase->obs[w].quality;
			msg->msg_type.type18.phaseobs[w].w4.continuityloss = (unsigned)tp->msg_data.phase->obs[w].continuityloss;
			dtmp = tp->msg_data.phase->obs[w].value;
			while (fabs(dtmp) > 8388608) {
				dtmp += (dtmp < 0) ? 8388608 : -8388608;
			}
			ltmp = (long)round(dtmp * 256.0);
			ltmp = ~ltmp;
			ltmp += 1;
			msg->msg_type.type18.phaseobs[w].w4.phase_h = (unsigned) ltmp >> 24;
			msg->msg_type.type18.phaseobs[w].w5.phase_l = (unsigned) ltmp & 0xffffff;
		}
	break;

	case 19:
		if (tp->msg_data.pseudo == NULL) break;
		msg->msg_type.type19.w3.gnsstime = (unsigned)tp->msg_data.gnsstime;
		msg->msg_type.type19.w3.smoothing = (unsigned)tp->msg_data.smth;
		msg->msg_type.type19.w3.frequency = (unsigned)tp->msg_data.pseudo->frequency;
		for (w = 0; w < tp->msg_data.numobs; w++) {
			msg->msg_type.type19.pseudoobs[w].w4.multiple = (unsigned)tp->msg_data.pseudo->obs[w].multiple;
			msg->msg_type.type19.pseudoobs[w].w4.cacode = (unsigned)tp->msg_data.pseudo->obs[w].cacode;
			msg->msg_type.type19.pseudoobs[w].w4.GPS = (unsigned)tp->msg_data.pseudo->obs[w].GPS;
			msg->msg_type.type19.pseudoobs[w].w4.prn = (unsigned)tp->msg_data.pseudo->obs[w].prn;
			msg->msg_type.type19.pseudoobs[w].w4.quality = (unsigned)tp->msg_data.pseudo->obs[w].quality;
			msg->msg_type.type19.pseudoobs[w].w4.multipath = (unsigned)tp->msg_data.pseudo->obs[w].multipath;
			itmp = (int) round(tp->msg_data.pseudo->obs[w].value / 0.02);
			msg->msg_type.type19.pseudoobs[w].w4.pseudo_h = (uint)itmp  >> 24;
			msg->msg_type.type19.pseudoobs[w].w5.pseudo_l = (uint)itmp & 0xffffff;
		}

	break;

	case 22:
		msg->msg_type.type22.w3.l1_x_delta = tp->msg_data.ecef_delta.l1_x_delta;
		msg->msg_type.type22.w3.l1_y_delta = tp->msg_data.ecef_delta.l1_y_delta;
		msg->msg_type.type22.w3.l1_z_delta = tp->msg_data.ecef_delta.l1_z_delta;
#if 0 /* This is in the spec, but if I add it in weird things happen */
		if (len >= 2) {
			tp->msg_data.ecef_delta.GPS = (bool) msg->msg_type.type22.w4.GPS;
			tp->msg_data.ecef_delta.at = (bool) msg->msg_type.type22.w4.at;
			tp->msg_data.ecef_delta.ar = (bool) msg->msg_type.type22.w4.ar;
			tp->msg_data.ecef_delta.nh = (bool) msg->msg_type.type22.w4.nh;
			tp->msg_data.ecef_delta.l1_phase_height = msg->msg_type.type22.w4.HI;
		}
		if (len >= 3) {
			tp->msg_data.ecef_delta.l2_x_delta = msg->msg_type.type22.w5.l2_x_delta;
			tp->msg_data.ecef_delta.l2_y_delta = msg->msg_type.type22.w5.l2_y_delta;
			tp->msg_data.ecef_delta.l2_z_delta = msg->msg_type.type22.w5.l2_z_delta;
		}
#endif
		break;

	case 23:
    	memset(ctmp, 0, sizeof(ctmp));
		n = 0;
		msg->msg_type.type23.w3.ar = (unsigned) tp->msg_data.antenna.reference;
		msg->msg_type.type23.w3.sf = (unsigned) tp->msg_data.antenna.serial;
		itmp = strlen(tp->msg_data.antenna.descriptor);
		msg->msg_type.type23.w3.nad = itmp;
		strcpy(ctmp, tp->msg_data.antenna.descriptor);
		n += itmp;

		ctmp[n++] = (int)tp->msg_data.antenna.setupid;
		itmp = (int)strlen(tp->msg_data.antenna.serialnum);
		ctmp[n++] = itmp;
		strcpy(&ctmp[n], tp->msg_data.antenna.serialnum);
		n += itmp;

		w = 0;
		msg->msg_type.type23.w3.byte1 = ctmp[w++]; n--;
		msg->msg_type.type23.w3.byte2 = ctmp[w++]; n--;
		while (n >= 3) {
			msg->msg_type.type23.info[itmp].byte1 = ctmp[w++];
			msg->msg_type.type23.info[itmp].byte2 = ctmp[w++];
			msg->msg_type.type23.info[itmp].byte3 = ctmp[w++];
			itmp++;
			n -= 3;
		}
		if (n == 2) {
			msg->msg_type.type23.info[itmp].byte1 = ctmp[w++];
			msg->msg_type.type23.info[itmp].byte2 = ctmp[w++];
			msg->msg_type.type23.info[itmp].byte3 = (char) 0xaa;
		}
		if (n == 1) {
			msg->msg_type.type23.info[itmp].byte1 = ctmp[w++];
			msg->msg_type.type23.info[itmp].byte2 = (char) 0xaa;
			msg->msg_type.type23.info[itmp].byte3 = (char) 0xaa;
		}

		break;

	case 24:
		memset(&u38tmp,0,sizeof(u38tmp));
		u38tmp = (long long int)(tp->msg_data.coord.ecef_x/.0001);
		u38tmp &= 0x3fffffffffLL; 			// set bits above 38 to 0
		msg->msg_type.type24.w3.ecef_x_h = u38tmp >> 14;
		msg->msg_type.type24.w4.ecef_x_l = u38tmp;

		memset(&u38tmp,0,sizeof(u38tmp));
		u38tmp = (long long int)(tp->msg_data.coord.ecef_y/.0001);
		u38tmp &= 0x3fffffffffLL; 			// set bits above 38 to 0
		msg->msg_type.type24.w4.ecef_y_h = u38tmp >> 30;
		msg->msg_type.type24.w5.ecef_y_m = u38tmp >> 6;
		msg->msg_type.type24.w6.ecef_y_l = u38tmp;

		memset(&u38tmp,0,sizeof(u38tmp));
		u38tmp = (long long int)(tp->msg_data.coord.ecef_z/.0001);
		u38tmp &= 0x3fffffffffLL; 			// set bits above 38 to 0
		msg->msg_type.type24.w6.ecef_z_h = u38tmp >> 22;
		msg->msg_type.type24.w7.ecef_z_l = u38tmp;

		msg->msg_type.type24.w7.haveHI = (uint) tp->msg_data.coord.haveHI;
		msg->msg_type.type24.w7.GPS = (uint) tp->msg_data.coord.GPS;
		msg->msg_type.type24.w8.HI = (int) (tp->msg_data.coord.HI *1e4);
		break;


    default:	/* U */
	memcpy(msg->msg_type.rtcm_msgunk, tp->msg_data.words, (RTCM_WORDS_MAX-2)*sizeof(isgps30bits_t));
	break;
    }

    /* compute parity for each word in the message */
    for (w = 0; w < tp->length; w++)
	wp[w].parity = isgps_parity(buf[w]);

    /* FIXME: must do inversion here */
    return true;
}

static bool preamble_match(isgps30bits_t *w)
{
    return (((struct rtcm_msghw1 *)w)->preamble == PREAMBLE_PATTERN);
}

static bool length_check(struct gps_packet_t *lexer)
{
    return lexer->isgps.bufindex >= 2 
	&& lexer->isgps.bufindex >= ((struct rtcm_msg_t *)lexer->isgps.buf)->w2.frmlen + 2u;
}

enum isgpsstat_t rtcm_decode(struct gps_packet_t *lexer, unsigned int c)
{
    return isgps_decode(lexer, 
			preamble_match, 
			length_check, 
			RTCM_WORDS_MAX, 
			c);
}

void rtcm_dump(struct rtcm_t *rtcm, /*@out@*/char buf[], size_t buflen)
/* dump the contents of a parsed RTCM104 message */
{
    unsigned int n,i;

    (void)snprintf(buf, buflen, "H\t%u\t%u\t%0.1f\t%u\t%u\t%u\n",
	   rtcm->type,
	   rtcm->refstaid,
	   rtcm->zcount,
	   rtcm->seqnum,
	   rtcm->length,
	   rtcm->stathlth);

    switch (rtcm->type) {
    case 1:
    case 9:
	for (n = 0; n < rtcm->msg_data.ranges.nentries; n++) {
	    struct rangesat_t *rsp = &rtcm->msg_data.ranges.sat[n];
	    (void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			   "S\t%u\t%u\t%u\t%0.1f\t%0.3f\t%0.3f\n",
			   rsp->ident,
			   rsp->udre,
			   rsp->issuedata,
			   rtcm->zcount,
			   rsp->rangerr,
			   rsp->rangerate);
	}
	break;

    case 3:
	if (rtcm->msg_data.ecef.valid)
	    (void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			   "R\t%.2f\t%.2f\t%.2f\n",
			   rtcm->msg_data.ecef.x, 
			   rtcm->msg_data.ecef.y,
			   rtcm->msg_data.ecef.z);
	break;

    case 4:
	if (rtcm->msg_data.reference.valid)
	    (void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			   "D\t%s\t%1d\t%s\t%.1f\t%.1f\t%.1f\n",
			   (rtcm->msg_data.reference.system==gps) ? "GPS"
			   : ((rtcm->msg_data.reference.system==glonass) ? "GLONASS"
			      : "UNKNOWN"),
			   rtcm->msg_data.reference.sense,
			   rtcm->msg_data.reference.datum,
			   rtcm->msg_data.reference.dx,
			   rtcm->msg_data.reference.dy,
			   rtcm->msg_data.reference.dz);
	break;

    case 5:
	for (n = 0; n < rtcm->msg_data.conhealth.nentries; n++) {
	    struct consat_t *csp = &rtcm->msg_data.conhealth.sat[n];
	    (void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			   "C\t%2u\t%1u\t%1u\t%2d\t%1u\t%1u\t%1u\t%2u\n",
			   csp->ident,
			   (unsigned)csp->iodl,
			   (unsigned)csp->health,
			   csp->snr,
			   (unsigned)csp->health_en,
			   (unsigned)csp->new_data,
			   (unsigned)csp->los_warning,
			   csp->tou);
	}
	break;

    case 6: 			/* NOP msg */
	(void)strlcat(buf, "N\n", buflen);
	break;

    case 7:
	for (n = 0; n < rtcm->msg_data.almanac.nentries; n++) {
	    struct station_t *ssp = &rtcm->msg_data.almanac.station[n];
	    (void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			   "A\t%.4f\t%.4f\t%u\t%.1f\t%u\t%u\t%u\n",
			   ssp->latitude,
			   ssp->longitude,
			   ssp->range,
			   ssp->frequency,
			   ssp->health,
			   ssp->station_id,
			   ssp->bitrate);
	}
	break;
    case 16:
	(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
		       "T\t\"%s\"\n", rtcm->msg_data.message);
	break;


    case 18:
	if (rtcm->msg_data.phase == NULL) break;
	(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"L\t%ld\t%d\t%d\n",
			rtcm->msg_data.gnsstime,
			rtcm->msg_data.phase->frequency,
			rtcm->msg_data.numobs);

	for (n = 0; n < rtcm->msg_data.numobs; n++) {
		(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
				"E\t%d\t%d\t%d\t%d\t%d\t%d\t%0.3lf\n",
				rtcm->msg_data.phase->obs[n].multiple,
				rtcm->msg_data.phase->obs[n].cacode,
				rtcm->msg_data.phase->obs[n].GPS,
				rtcm->msg_data.phase->obs[n].prn,
				rtcm->msg_data.phase->obs[n].quality,
				rtcm->msg_data.phase->obs[n].continuityloss,
				rtcm->msg_data.phase->obs[n].value);
	}
	break;

    case 19:
	if (rtcm->msg_data.pseudo == NULL) break;
	(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"P\t%ld\t%d\t%d\t%d\n",
			rtcm->msg_data.gnsstime,
			rtcm->msg_data.smth,
			rtcm->msg_data.pseudo->frequency,
			rtcm->msg_data.numobs);

	for (n = 0; n < rtcm->msg_data.numobs; n++) {
		(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
				"O\t%d\t%d\t%d\t%d\t%d\t%d\t%0.3lf\n",
				rtcm->msg_data.pseudo->obs[n].multiple,
				rtcm->msg_data.pseudo->obs[n].cacode,
				rtcm->msg_data.pseudo->obs[n].GPS,
				rtcm->msg_data.pseudo->obs[n].prn,
				rtcm->msg_data.pseudo->obs[n].quality,
				rtcm->msg_data.pseudo->obs[n].multipath,
				rtcm->msg_data.pseudo->obs[n].value);
	}
	break;

    case 22:
	(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"M1\t%0.9lf\t%0.9lf\t%0.9lf\n",
			(double)rtcm->msg_data.ecef_delta.l1_x_delta/256.0,
			(double)rtcm->msg_data.ecef_delta.l1_y_delta/256.0,
			(double)rtcm->msg_data.ecef_delta.l1_z_delta/256.0);
	if (rtcm->msg_data.ecef_delta.haveHI) {
		
		double HI = (double)rtcm->msg_data.ecef_delta.l1_phase_height/256.0;
		if (! rtcm->msg_data.ecef_delta.nh)
			HI = 0.0;
		(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"MH\t%d\t%d\t%d\t%d\t%lf\n",
			rtcm->msg_data.ecef_delta.GPS,
			rtcm->msg_data.ecef_delta.at,
			rtcm->msg_data.ecef_delta.ar,
			rtcm->msg_data.ecef_delta.nh,
			HI);
	}
	if (rtcm->msg_data.ecef_delta.haveL2) {
		(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"M2\t%0.9lf\t%0.9lf\t%0.9lf\n",
			(double)rtcm->msg_data.ecef_delta.l2_x_delta/256.0,
			(double)rtcm->msg_data.ecef_delta.l2_y_delta/256.0,
			(double)rtcm->msg_data.ecef_delta.l2_z_delta/256.0);
	}
	break;

    case 23:
	(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"I\t%d\t%d\t%u\t%u\t%u\n",
			rtcm->msg_data.antenna.reference,
			rtcm->msg_data.antenna.serial,
			rtcm->msg_data.antenna.setupid,
			rtcm->msg_data.antenna.len_desc,
			rtcm->msg_data.antenna.len_serial);
   	(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"ID\t%s\n",
			rtcm->msg_data.antenna.descriptor);
   	(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"IS\t%s\n",
			rtcm->msg_data.antenna.serialnum);
	break;

    case 24:
	(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"Z\t%d\t%d\t%lf\t%lf\t%lf\t%lf\n",
			rtcm->msg_data.coord.GPS,
			rtcm->msg_data.coord.haveHI,
			rtcm->msg_data.coord.ecef_x,
			rtcm->msg_data.coord.ecef_y,
			rtcm->msg_data.coord.ecef_z,
			rtcm->msg_data.coord.HI);
	break;
    case 1001: case 1002: case 1003: case 1004:
	(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"1\t%ld\t%d\t%d\t%d\t%d\n",
			rtcm->msg_data.gpstow,
			rtcm->msg_data.gnss_sync,
			rtcm->msg_data.numobs,
			rtcm->msg_data.smth,
			rtcm->msg_data.smth_int);


	for (i = 0; i < RTCMMAXTYPES; i++) {
		if (! rtcm->msg_data.obs_data[i].obs_avail) continue;

		for (n = 0; n < rtcm->msg_data.numobs; n++) {
			(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
					"11\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%0.5lf\n", i,
					rtcm->msg_data.obs_data[i].obs[n].multiple,
					rtcm->msg_data.obs_data[i].obs[n].cacode,
					rtcm->msg_data.obs_data[i].obs[n].GPS,
					rtcm->msg_data.obs_data[i].obs[n].prn,
					rtcm->msg_data.obs_data[i].obs[n].quality,
					rtcm->msg_data.obs_data[i].obs[n].continuityloss,
					rtcm->msg_data.obs_data[i].obs[n].multipath,
					rtcm->msg_data.obs_data[i].obs[n].locktime,
					rtcm->msg_data.obs_data[i].obs[n].value);
		}
	}
	break;
    case 1005: case 1006:
	(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"5\t%d\t%d\t%u\t%u\t%u\n",
			rtcm->msg_data.antenna.reference,
			rtcm->msg_data.antenna.serial,
			rtcm->msg_data.antenna.setupid,
			rtcm->msg_data.antenna.len_desc,
			rtcm->msg_data.antenna.len_serial);
	(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"51\t%d\t%d\t%lf\t%lf\t%lf\n",
			rtcm->msg_data.coord.GPS,
			rtcm->msg_data.coord.haveHI,
			rtcm->msg_data.coord.ecef_x,
			rtcm->msg_data.coord.ecef_y,
			rtcm->msg_data.coord.ecef_z);
        if (rtcm->type == 1006) {
	    (void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"62\t%lf\n",
			rtcm->msg_data.coord.HI);
      
        }
	break;
    case 1007: case 1008:
   	(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"7\t%d\t%s\n",
			rtcm->msg_data.antenna.setupid,
			rtcm->msg_data.antenna.descriptor);
        if (rtcm->type == 1008) {
   	    (void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			    "81\t%s\n",
			    rtcm->msg_data.antenna.serialnum);
        }
	break;

    default:
	for (n = 0; n < rtcm->length; n++)
	    (void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			   "U\t0x%08x\n", rtcm->msg_data.words[n]);
	break;
    }

}

int rtcm_undump(/*@out@*/struct rtcm_t *rtcmp, char *buf)
/* merge a line of data into an RTCM structure, return 0 if done */
{
    int fldcount, v;
    unsigned n;
    char buf2[BUFSIZ]; /* stdio.h says BUFSIZ=1024. True everywhere? */
    static int loopcount;
    uint a,b,c;
    double da,db,dc;
    int i;


    /*@ -usedef @*/
    switch (rtcmp->type) {
    case 0:
	fldcount = sscanf(buf, "H\t%u\t%u\t%lf\t%u\t%u\t%u\n",
			&rtcmp->type,
			&rtcmp->refstaid,
			&rtcmp->zcount,
			&rtcmp->seqnum,
			&rtcmp->length,
			&rtcmp->stathlth);
	if (fldcount != 6)
	    return -1;
	else
	    return 1;
	//break;

    case 1:
    case 9:
	{
	    struct rangesat_t *rsp = &rtcmp->msg_data.ranges.sat[rtcmp->msg_data.ranges.nentries++];
	    /* we ignore the third (zcount) field, it's in the parent */
	    fldcount = sscanf(buf,
			      "S\t%u\t%u\t%u\t%*f\t%lf\t%lf\n",
			      &rsp->ident,
			      &rsp->udre,
			      &rsp->issuedata,
			      &rsp->rangerr,
			      &rsp->rangerate);
	    if (fldcount != 5 || (rtcmp->type != 1 && rtcmp->type != 9))
		return (int)(-rtcmp->type-1);
	    else if (rtcmp->msg_data.ranges.nentries != rtcmp->length*3/5)
		return (int)(rtcmp->type+1);
	    else
		return 0;
	}
	//break;

    case 3:
	fldcount = sscanf(buf,
			  "R\t%lf\t%lf\t%lf\n",
			  &rtcmp->msg_data.ecef.x, 
			  &rtcmp->msg_data.ecef.y,
			  &rtcmp->msg_data.ecef.z);
	if (fldcount != 3 || rtcmp->type != 3)
	    return -4;
	else {
	    rtcmp->msg_data.ecef.valid = true;
	    return 0;
	}
	//break;

    case 4:
	fldcount = sscanf(buf,
			   "D\t%1023s\t%1d\t%5s\t%lf\t%lf\t%lf\n",
			  buf2,
			  &v,
			  (char *)&rtcmp->msg_data.reference.datum,
			  &rtcmp->msg_data.reference.dx,
			  &rtcmp->msg_data.reference.dy,
			  &rtcmp->msg_data.reference.dz);
	if (fldcount != 6 || rtcmp->type != 4)
	    return -5;
	else {
	    if (strcmp(buf2, "GPS") == 0)
		rtcmp->msg_data.reference.system = gps;
	    else if (strcmp(buf2, "GLONASS") == 0)
		rtcmp->msg_data.reference.system = glonass;
	    else
		rtcmp->msg_data.reference.system = unknown;
	    rtcmp->msg_data.reference.sense = (v == 1) ? global : ((v == 0) ? local : invalid);
	    rtcmp->msg_data.reference.valid = true;
	    return 0;
	}
	//break;

    case 5:
	{
	    struct consat_t *csp = &rtcmp->msg_data.conhealth.sat[rtcmp->msg_data.conhealth.nentries++];
	    unsigned int iodl, new_data, los_warning;

	    fldcount = sscanf(buf,
			      "C\t%2u\t%1u\t%1u\t%2d\t%1u\t%1u\t%1u\t%2u\n",
			      &csp->ident,
			      &iodl,
			      &csp->health,
			      &csp->snr,
			      &csp->health_en,
			      &new_data,
			      &los_warning,
			      &csp->tou);
	    csp->iodl = iodl > 0;
	    csp->new_data = new_data > 0;
	    csp->los_warning = los_warning > 0;
	    if (fldcount != 8 || rtcmp->type != 5)
		return -6;
	    else if (rtcmp->msg_data.conhealth.nentries < rtcmp->length)
		return 6;
	    else
		return 0;
	}
	//break;

    case 6: 			/* NOP msg */
	if (buf[0] != 'N')
	    return -7;
	else
	    return 0;
	//break;

    case 7:
	{
	    struct station_t *ssp = &rtcmp->msg_data.almanac.station[rtcmp->msg_data.almanac.nentries++];
	    fldcount = sscanf(buf,
			      "A\t%lf\t%lf\t%u\t%lf\t%u\t%u\t%u\n",
			      &ssp->latitude,
			      &ssp->longitude,
			      &ssp->range,
			      &ssp->frequency,
			      &ssp->health,
			      &ssp->station_id,
			      &ssp->bitrate);
	    if (fldcount != 7 || rtcmp->type != 7)
		return 8;
	    else if (rtcmp->msg_data.almanac.nentries < rtcmp->length/3)
		return 8;
	    else
		return 0;
	}
	//break;

    case 16:
	fldcount = sscanf(buf, "T\t\"%[^\"]\"\n", rtcmp->msg_data.message);
	if (fldcount != 1)
	    return 16;
	else
	    return 0;
	//break;

    case 18:

	if (buf[0] == 'L') {
		unsigned long a;
		unsigned int freq,c;
		fldcount = sscanf(buf, "L\t%lu\t%u\t%u\n",
				&a, &freq, &c);
		int s = (freq == 0) ? RTCM_L1 : RTCM_L2;
		rtcmp->msg_data.phase = &(rtcmp->msg_data.obs_data[s]);
		rtcmp->msg_data.gnsstime = a;
		rtcmp->msg_data.phase->frequency = freq;
		rtcmp->msg_data.numobs = c;
		loopcount = 0;
		return(18);
	} else if (buf[0] == 'E') {
		uint a,b,c;

		if (rtcmp->msg_data.phase == NULL) return(-18);
		fldcount = sscanf(buf, "E\t%u\t%u\t%u\t%u\t%u\t%u\t%lf\n",
				&a, &b, &c,
				&rtcmp->msg_data.phase->obs[loopcount].prn,
				&rtcmp->msg_data.phase->obs[loopcount].quality,
				&rtcmp->msg_data.phase->obs[loopcount].continuityloss,
				&rtcmp->msg_data.phase->obs[loopcount].value);
		rtcmp->msg_data.phase->obs[loopcount].multiple = (bool)a;
		rtcmp->msg_data.phase->obs[loopcount].cacode = (bool)b;
		rtcmp->msg_data.phase->obs[loopcount].GPS = (bool)c;
		loopcount++;
		if (loopcount < rtcmp->msg_data.numobs) return(18);
		// loopcount is done
		rtcmp->msg_data.phase->obs_avail = true;
		return(0);
	} else {
		return (-18);
	}

	break;

    case 19:
	if (rtcmp->msg_data.pseudo == NULL)
		rtcmp->msg_data.pseudo = &(rtcmp->msg_data.obs_data[0]);
	if (buf[0] == 'P') {
		fldcount = sscanf(buf, "P\t%lu\t%d\t%u\t%u\n",
				&rtcmp->msg_data.gnsstime,
				&a,
				&rtcmp->msg_data.pseudo->frequency,
				&rtcmp->msg_data.numobs);
		rtcmp->msg_data.smth = (bool)a;
		loopcount = 0;
		return(19);
	} else if (buf[0] == 'O') {
		fldcount = sscanf(buf, "O\t%u\t%u\t%u\t%u\t%u\t%u\t%lf\n",
				&a, &b, &c,
				&rtcmp->msg_data.pseudo->obs[loopcount].prn,
				&rtcmp->msg_data.pseudo->obs[loopcount].quality,
				&rtcmp->msg_data.pseudo->obs[loopcount].multipath,
				&rtcmp->msg_data.pseudo->obs[loopcount].value);
		rtcmp->msg_data.pseudo->obs[loopcount].multiple = (bool)a;
		rtcmp->msg_data.pseudo->obs[loopcount].cacode = (bool)b;
		rtcmp->msg_data.pseudo->obs[loopcount].GPS = (bool)c;
		loopcount++;
		if (loopcount < rtcmp->msg_data.numobs) return(19);
		// loopcount is done, set pointer

		int s = RTCM_C1;
		if (rtcmp->msg_data.phase->obs[0].cacode == 0) {
			s = (rtcmp->msg_data.phase->frequency == 0) ? RTCM_P1 : RTCM_P2;
		}
		if (s != 0) {
			memcpy(&(rtcmp->msg_data.obs_data[s]), rtcmp->msg_data.pseudo, sizeof(rtcmp->msg_data.obs_data[s]));
			rtcmp->msg_data.pseudo = &(rtcmp->msg_data.obs_data[s]);
		}
		rtcmp->msg_data.pseudo->obs_avail = true;
		return(0);
	} else {
		return(-19);
	}

	break;

    case 22:
	switch (buf[1]) {
		case '1' :
		fldcount = sscanf(buf, "M1\t%lf\t%lf\t%lf\n",&da,&db,&dc);
		rtcmp->msg_data.ecef_delta.l1_x_delta = da * 256;
		rtcmp->msg_data.ecef_delta.l1_y_delta = db * 256;
		rtcmp->msg_data.ecef_delta.l1_z_delta = dc * 256;
		return(0);
		break;
	}
#if 0 /* This is in the spec, but if I add it in weird things happen */
	case 'H':
		
		double HI = (double)session->gpsdata.rtcm.msg_data.ecef_delta.l1_phase_height/256.0;
		if (! session->gpsdata.rtcm.msg_data.ecef_delta.nh)
			HI = 0.0;
		(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"MH\t%d\t%d\t%d\t%d\t%lf\n",
			session->gpsdata.rtcm.msg_data.ecef_delta.GPS,
			session->gpsdata.rtcm.msg_data.ecef_delta.at,
			session->gpsdata.rtcm.msg_data.ecef_delta.ar,
			session->gpsdata.rtcm.msg_data.ecef_delta.nh,
			HI);
		case '2':
		(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
			"M2\t%lf\t%lf\t%lf\n",
			(double)session->gpsdata.rtcm.msg_data.ecef_delta.l2_x_delta/256.0,
			(double)session->gpsdata.rtcm.msg_data.ecef_delta.l2_y_delta/256.0,
			(double)session->gpsdata.rtcm.msg_data.ecef_delta.l2_z_delta/256.0);
#endif
	break;

    case 23:
	switch (buf[1]) {
		case '\t': 
			fldcount = sscanf(buf, "I\t%d\t%d\t%u\t%u\t%u\n",
				&a, &b,
				&rtcmp->msg_data.antenna.setupid,
				&rtcmp->msg_data.antenna.len_desc,
				&rtcmp->msg_data.antenna.len_serial);
			rtcmp->msg_data.antenna.reference = (bool) a;
			rtcmp->msg_data.antenna.serial = (bool) b;
			return(23);
		case 'D':
   			fldcount = sscanf(buf, "ID\t%[^\n]\n",
				rtcmp->msg_data.antenna.descriptor);
			for (i = 0; i < sizeof(rtcmp->msg_data.antenna.descriptor); i++) {
				if (!isprint(rtcmp->msg_data.antenna.descriptor[i])) {
					rtcmp->msg_data.antenna.descriptor[i] = '\0';
					break;
				}
			}
			return(23);
		case 'S':
   			fldcount = sscanf(buf, "IS\t%[^\n]\n",
				rtcmp->msg_data.antenna.serialnum);
			for (i = 0; i < sizeof(rtcmp->msg_data.antenna.serialnum); i++) {
				if (!isprint(rtcmp->msg_data.antenna.serialnum[i])) {
					rtcmp->msg_data.antenna.serialnum[i] = '\0';
					break;
				}
			}
			return(0);
	}
	break;

    case 24:
	fldcount = sscanf(buf, "Z\t%d\t%d\t%lf\t%lf\t%lf\t%lf\n",
			&a, &b,
			&rtcmp->msg_data.coord.ecef_x,
			&rtcmp->msg_data.coord.ecef_y,
			&rtcmp->msg_data.coord.ecef_z,
			&rtcmp->msg_data.coord.HI);
	rtcmp->msg_data.coord.GPS = (bool) a;
	rtcmp->msg_data.coord.haveHI = (bool) b;
	if (fldcount == 6)
		return(0);
	else
		return(-24);
	break;

    case 1001: case 1002: case 1003: case 1004:
	switch (buf[1]) {
#if 0
		case '\t':
			fldcount = (void)sscanf(buf, "1\t%ld\t%d\t%d\t%d\t%d\n",
					&rtcmp->msg_data.gpstow,
					&rtcmp->gnss_sync,
					&rtcmp->msg_data.numobs,
					&rtcmp->smth,
					&rtcmp->smth_int);

			break;
		case '1':
			(void)snprintf(buf + strlen(buf), buflen - strlen(buf),
					"11\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%0.3lf\n", i,
					rtcm->msg_data.obs_data[i].obs[n].multiple,
					rtcm->msg_data.obs_data[i].obs[n].cacode,
					rtcm->msg_data.obs_data[i].obs[n].GPS,
					rtcm->msg_data.obs_data[i].obs[n].prn,
					rtcm->msg_data.obs_data[i].obs[n].quality,
					rtcm->msg_data.obs_data[i].obs[n].continuityloss,
					rtcm->msg_data.obs_data[i].obs[n].multipath,
					rtcm->msg_data.obs_data[i].obs[n].locktime,
					rtcm->msg_data.obs_data[i].obs[n].value);
			break;
#endif
		break;
	}
	break;
    case 1005: case 1006:
	switch (buf[1]) {
		case '\t':
			fldcount = sscanf(buf, "5\t%d\t%d\t%u\t%u\t%u\n",
					&a, &b,
					&rtcmp->msg_data.antenna.setupid,
					&rtcmp->msg_data.antenna.len_desc,
					&rtcmp->msg_data.antenna.len_serial);
			rtcmp->msg_data.antenna.reference = (a) ? true : false;
			rtcmp->msg_data.antenna.serial = (b) ? true : false;
			return(rtcmp->type);
		case '1':
			(void)sscanf(buf, "51\t%d\t%d\t%lf\t%lf\t%lf\n",
					&a, &b,
					&rtcmp->msg_data.coord.ecef_x,
					&rtcmp->msg_data.coord.ecef_y,
					&rtcmp->msg_data.coord.ecef_z);
			rtcmp->msg_data.coord.GPS = (a) ? true : false;
			rtcmp->msg_data.coord.haveHI = (b) ? true : false;
			if (rtcmp->type == 1006) return (1006);
			return(0);
		case '2':
			(void)sscanf(buf, "62\t%lf\n",
					&rtcmp->msg_data.coord.HI);
			return(0);
      
        }
	break;
    case 1007: case 1008:
	switch (buf[1]) {
		case '\t':
   			fldcount = sscanf(buf, "7\t%d\t%[^\n]\n",
					&rtcmp->msg_data.antenna.setupid,
					rtcmp->msg_data.antenna.descriptor);
			for (i = 0; i < sizeof(rtcmp->msg_data.antenna.descriptor); i++) {
				if (!isprint(rtcmp->msg_data.antenna.descriptor[i])) {
					rtcmp->msg_data.antenna.descriptor[i] = '\0';
					break;
				}
			}
			if (rtcmp->type == 1008) return(1008);
			return(0);
		case '1':
   			fldcount = sscanf(buf, "81\t%[^\n]\n",
					rtcmp->msg_data.antenna.serialnum);
			for (i = 0; i < sizeof(rtcmp->msg_data.antenna.serialnum); i++) {
				if (!isprint(rtcmp->msg_data.antenna.serialnum[i])) {
					rtcmp->msg_data.antenna.serialnum[i] = '\0';
					break;
				}
			}
			return(0);
	}
	break;

    default:
	for (n = 0; n < DIMENSION(rtcmp->msg_data.words); n++)
	    if (rtcmp->msg_data.words[n] == 0)
		break;
	if (n >= DIMENSION(rtcmp->msg_data.words))
	    return 0;
	else {
	    fldcount = sscanf(buf, "U\t0x%08x\n", &v);
	    if (fldcount != 1)
		return (int)(-rtcmp->type-1);
	    else {
		rtcmp->msg_data.words[n] = (isgps30bits_t)v;
		if (n == rtcmp->length-1)
		    return 0;
		else
		    return (int)(rtcmp->type+1);
	    }
	}
	//break;
    }
    /*@ +usedef @*/
    return(-1);
}

/*
 * The RTCM words are 30-bit words.  We will lay them into memory into
 * 30-bit (low-end justified) chunks.  To write them out we will write
 * 5 Magnavox-format bytes where the low 6-bits of the byte are 6-bits
 * of the 30-word msg.
 */
#define P_30_MASK       0x40000000u
#define MAG_SHIFT 6u
#define MAG_TAG_DATA (1 << MAG_SHIFT)
#define W_DATA_MASK     0x3fffffc0u
unsigned int isgps_parity(isgps30bits_t);
static unsigned int reverse_bits[] = {
    0, 32, 16, 48, 8, 40, 24, 56, 4, 36, 20, 52, 12, 44, 28, 60,
    2, 34, 18, 50, 10, 42, 26, 58, 6, 38, 22, 54, 14, 46, 30, 62,
    1, 33, 17, 49, 9, 41, 25, 57, 5, 37, 21, 53, 13, 45, 29, 61,
    3, 35, 19, 51, 11, 43, 27, 59, 7, 39, 23, 55, 15, 47, 31, 63
};


/*
 * Mixing calls to rtcm_output_magnavox and rtcm_buffer_magnavox will 
 * result in a ISGPS parity failure method because of static variables
 */
void rtcm_output_magnavox(isgps30bits_t * ip, FILE *fp)
/* ship an RTCM message to standard output in Magnavox format */
{
    static isgps30bits_t w = 0;
    int             len;
    static uint     sqnum = 0;

    struct rtcm_msg_t  *msg = (struct rtcm_msg_t *) ip;

    len = msg->w2.frmlen+2;
    msg->w2.sqnum = sqnum++;
    sqnum &= 0x7;

    while (len-- > 0) {
	w <<= 30;
	w |= *ip++ & W_DATA_MASK;

	w |= isgps_parity(w);

	/* weird-assed inversion */
	if (w & P_30_MASK)
	    w ^= W_DATA_MASK;

	/* msb first */
	putchar(MAG_TAG_DATA | reverse_bits[(w >> 24) & 0x3f]);
	putchar(MAG_TAG_DATA | reverse_bits[(w >> 18) & 0x3f]);
	putchar(MAG_TAG_DATA | reverse_bits[(w >> 12) & 0x3f]);
	putchar(MAG_TAG_DATA | reverse_bits[(w >> 6) & 0x3f]);
	putchar(MAG_TAG_DATA | reverse_bits[(w) & 0x3f]);
    }
//	fflush(stdout);
}

void rtcm_buffer_magnavox(isgps30bits_t * ip, unsigned char *buf, unsigned int *offset)
/* ship an RTCM message to standard output in Magnavox format */
{
    int             len;
    static uint     sqnum = 0;

    struct rtcm_msg_t  *msg = (struct rtcm_msg_t *) ip;
    len = msg->w2.frmlen+2;
    msg->w2.sqnum = sqnum++;
    sqnum &= 0x7;

    isgps_buffer_magnavox(ip, len, buf, offset);
}

#endif /* RTCM104_ENABLE */
